// Copyright 2016 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

//go:generate go run gen.go -mdicons=/path/to/the/material-design-icons

// Package icons contains the Material Design icon set, in the IconVG vector
// graphic format.
//
// See https://design.google.com/icons/ and
// https://godoc.org/golang.org/x/exp/shiny/iconvg
package icons
